package gui.tasksetgui;

import javax.swing.BoxLayout;
import javax.swing.JPanel;
import java.awt.Dimension;
import java.util.Observer;
import java.util.Observable;
import java.util.Optional;
import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.NoSuchElementException;

import tasks.Taskset;
import gui.BoxPanel;
import gui.UpdateListener;

public class TasksetDisplay extends BoxPanel implements Observer, UpdateListener, MarkListener, EmptyListener {

    private final int offset = 200;
    private Taskset taskset;
    private List<TaskGrid> gridList = new ArrayList<>();

    private TaskGridHeader header = new TaskGridHeader(" Priority", "Exectime", " Period", "Deadline");
    private Optional<EmptyTaskGrid> empty = Optional.empty();
    private Optional<TaskGrid> marked = Optional.empty(); // Used to indicate which taskgrid is marked


    public TasksetDisplay(Taskset taskset) {
        super(BoxLayout.PAGE_AXIS);
        this.taskset = taskset;
        taskset.addObserver(this);

        add(header);
        setMaximumSize(new Dimension(300, 35));
        add(createVerticalStrut(offset));
    }

    public void update(Observable obs, Object obj) {
        //Updates the box list
        removeAll();
        gridList.clear();
        add(header);

        Set<Integer> priorities = taskset.getPriorities();
        for (int i : priorities) {
            int[] t = taskset.getTask(i).get().attributes();
            TaskGrid tg = new TaskGrid(parse(i), parse(t[0]), parse(t[1]), parse(t[2]));
            tg.addUpdateListener(this);
            tg.addMarkListener(this);

            add(tg);
            gridList.add(tg);

            if (marked.map(taskgrid -> taskgrid.equals(tg)).orElse(false)) marked = Optional.of(tg);

            revalidate();
        }

        // This section fixes the offset from 
        // the lowest priority task to the Southern border.
        int taskOffset = priorities.size()*35;
        if (offset - taskOffset > 0) {
            setMaximumSize(new Dimension(300, 35 + taskOffset));
            add(createVerticalStrut(offset - taskOffset));
            revalidate();
        }

        // Mark the previously marked one.
        marked.ifPresent(taskGrid -> taskGrid.mark());

        // Add empty if exists
        if (empty.isPresent()) {
            add(empty.get());
            revalidate();
        }
    }

    /*
     * The function to be performed when the observable class is updated.
     * NOTE: TWO INTERFACES ARE NEEDED SUCH THAT THE UI ISN'T UPDATED ONLY WHEN MARKED.
     */
    public void update(Object obj) {
        TaskGrid tg = (TaskGrid) obj;
        int[] attributes = tg.getAttributes();

        try {
            taskset.changeTask(attributes[0], attributes[1], attributes[2], attributes[3]);
        } catch (NoSuchElementException e) {}
    }

    /*
     * The function to be performed when the observable class is marked. 
     * NOTE: TWO INTERFACES ARE NEEDED SUCH THAT THE UI ISN'T UPDATED ONLY WHEN MARKED.
     */
    public void markUpdate(Object obj) {
        marked = Optional.of((TaskGrid) obj);

        for (TaskGrid tg : gridList) {
            //if (marked.map(taskgrid -> taskgrid.equals(tg)).orElse(false)) marked = Optional.of(tg);
            tg.unmark();
        }

        marked.get().mark();
    }

    /*
     * Updates the empty grid when things happen
     */
    public void emptyUpdate() {
        if (empty.isPresent() && !empty.get().isEmpty()) {
            int[] attr = empty.get().getAttributes();
            empty = Optional.empty();
            taskset.addTask(attr[1], attr[2], attr[3]);
        }
    }

    /*
     * Adds EmptyTaskGrid unless one already exists
     */
    public void addEmpty() {
        if (!empty.isPresent()) {
            String prio = parse(taskset.nbrTasks() + 1);

            EmptyTaskGrid emptyGrid = new EmptyTaskGrid(prio);
            emptyGrid.addEmptyListener(this);
            empty = Optional.of(emptyGrid);

            add(emptyGrid);
            revalidate();
        }
    }
    
    /*
     * Removes marked task. Not if none is marked.
     */
    public void removeMarked() {
        int prio = marked
                    .map(grid -> grid.getAttributes()[0])
                    .orElse(-1);

        try {
            if (prio != -1) {
                taskset.removeTask(prio);
                marked = Optional.empty();
            }
        } catch (NoSuchElementException e) {}
    }

    // Private help method.
    private String parse(int i) {
        return String.valueOf(i);
    }
}
