package model;

import java.util.Observable;
import java.util.Observer;
import java.util.Map;
import java.util.HashMap;
import java.util.Arrays;

import tasks.Taskset;
import strategy.SchedulingStrategy;
import strategy.NoneStrategy;

public class SchedulingModel extends Observable implements Observer {
    private Schedule solution;
    private Taskset taskset;
    private Status status;
    private Map<String, Cell> cells;

    public SchedulingModel(Taskset ts, Status status) {
        taskset = ts;
        taskset.addObserver(this);
        this.status = status;
        solution = new Schedule(taskset);
        cells = new HashMap<>();
    }

    /*
     * Change Scheduling strategy
     */
    public void setStrategy(SchedulingStrategy strategy) {
        solution.setStrategy(strategy);
        taskset.setComparator(strategy.prioComparator()); // Sets the comparator which orders the tasks
        status.setStatus("Strategy Changed To " + strategy);

        setChanged();
        notifyObservers();
    }

    /*
     * Get schedule as an int vector, each pos represent task number
     */
    public int[] getSolution() {
        return solution.generateSchedule();
    }

    /*
     * returns whether solution matches the schedule generated by the user
     */
    public boolean match() {
        int[] sol = solution.generateSchedule();
        if (cells.size() != countNonZero(sol)) return false; // Checks if have the same number of elements in both

        int[] userSol = new int[sol.length];
        Arrays.fill(userSol, -1);
        for (Map.Entry<String, Cell> entry : cells.entrySet()) {
            String id = entry.getKey();
            int task = parseTask(id);
            int time = parseTime(id) - 1; // -1 to fit 0 based arrays

            if (userSol[time] != -1 || task != sol[time]) return false;
            userSol[time] = task;
        }

        return true;
    }

    // Parse + help functions
    private int countNonZero(int[] arr) {
        int n = 0;
        for (int i : arr) n = (i != 0) ? n+1 : n;
        return n;
    }
    private int parseTask(String id) { return Integer.parseInt(id.trim().split(",")[0].substring(1)); }
    private int parseTime(String id) {
        String time = id.trim().split(",")[1];
        return Integer.parseInt(time.substring(0, time.length() - 1));
    }

    /*
     * returns whether the schedule solution is valid
     */
    public boolean solutionIsValid() {
        return solution.isValid();
    }

    /*
     * Remove cell with id from this model.
     */
    private void clearCell(String id) {
        cells.remove(id);
        status.setStatus("Cell " + id + " cleared");
        setChanged();
        notifyObservers();
    }
    
    /*
     * Remove all cells from this model.
     */
    public void clearAllCells() {
        cells.clear();
        status.setStatus("All cells cleared");
        setChanged();
        notifyObservers();
    }

    /*
     * Remove all cells from this model.
     */
    public void clearAll() {
        cells.clear();                      // Clear all cells
        int size = taskset.nbrTasks();           // Remove all tasks
        for (int i = 0; i < size; i++) {
            try {
                taskset.removeTask(1);
            } catch (Exception e) {}
        }
        solution.setStrategy(new NoneStrategy()); // Change strategy to None

        status.setStatus("ScheduLearn model cleared");
        setChanged();
        notifyObservers();
    }
    
    /*
     * Change cell with id to coloredCell based on whether it was activated or deactivated.
     */
    public void activateCell(String id, int task, boolean activated) {
        String str = (activated) ? "activated" : "deactivated";

        if (!activated) clearCell(id);
        else cells.put(id, new ColorCell(task));

        status.setStatus("Cell " + id + " " + str);
        setChanged();
        notifyObservers();
    }
    /*
     * Return whether cell with string id exists or not.
     */
    public boolean cellExists(String id) {
        return cells.containsKey(id);
    }

    public boolean isCleared() { return cells.isEmpty(); }

    /*
     * Returns the value of cell with id as a string.
     */
    public String cellToString(String id) {
        if (!cells.containsKey(id)) return "";
        return cells.get(id).toString();
    }

    /*
     * When taskset is updated, Solution is as well
     */
    public void update(Observable obs, Object obj) {
        SchedulingStrategy strategy = solution.getStrategy();
        taskset = (Taskset)obs;

        solution = new Schedule(taskset);
        solution.setStrategy(strategy);

        // IF STRATEGY != PREV_STRATEGY DO:
        // Sets the taskset comparator which orders the tasks only if:
        // it wasn't the set comparator function which called update.
        if (obj == null) taskset.setComparator(strategy.prioComparator());

        status.setStatus("Strategy Changed To " + strategy);

        setChanged();
        notifyObservers();
    }
}
