package tasks;

import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.TreeSet;
import java.util.Observable;
import java.util.Optional;
import java.util.Comparator;
import java.util.NoSuchElementException;
import java.io.FileNotFoundException;

public class Taskset extends Observable {
    private List<Task> tasks;
    private Task idle;
    private Comparator<Task> stratComp;
    
    /*
     * Constructor
     */
    public Taskset() {
        tasks = new ArrayList<>();
    }

    /*
     * Add task, with attributes, to taskset
     */
    public void addTask(int exec, int period, int deadline) {
        tasks.add(new Task(exec, period, deadline));
        tasks.sort(stratComp); // Sort tasks in order according to comparator
        idle = generateIdle();

        setChanged();
        notifyObservers();
    }

    /*
     * Sorts this Taskset based on the strategy comparator
     */
    public void setComparator(Comparator<Task> stratComp) {
        this.stratComp = stratComp;
        tasks.sort(stratComp); // Sort tasks in order according to comparator

        setChanged();
        notifyObservers(new Boolean(true));
    }
    
    /*
     * change task, with attributes, in taskset
     */
    public void changeTask(int prio, int exec, int period, int deadline) throws NoSuchElementException {
        if (tasks.size() < prio) throw new NoSuchElementException("Task with priority " + prio + " doesn't exist!");
        
        tasks.set(prio - 1, new Task(exec, period, deadline));
        tasks.sort(stratComp); // Sort tasks in order according to comparator
        idle = generateIdle();

        setChanged();
        notifyObservers();
    }

    /*
     * Remove task with priority from taskset
     */
    public void removeTask(int prio) throws NoSuchElementException {
        if (tasks.size() < prio) throw new NoSuchElementException("Task with priority " + prio + " doesn't exist!");
        
        tasks.remove(prio - 1);
        tasks.sort(stratComp); // Sort tasks in order according to comparator
        idle = generateIdle();

        setChanged();
        notifyObservers();
    }

    /*
     * Generates new Idle task based on current tasks in taskset
     */
    private Task generateIdle() {
        double idle_u = 1 - utilization();
        int idle_period = hyperperiod();
        int idle_exectime = (int) (idle_u * idle_period);
        return new Task(idle_exectime, idle_period, idle_period);
    }

    /*
     * Prints the taskset by printing:
     * Each task with: Priority, execution time, period, and deadline.
     */
    public String toString() {
        StringBuilder sb = new StringBuilder("Pri\te_i\tt_i\td_i\n");
        for (int i = 0; i < tasks.size(); i++) {
            Task t = tasks.get(i);
            sb.append(i + "\t"
                    + t.exectime() + "\t" 
                    + t.period() + "\t" 
                    + t.deadline() + "\n");
        }
        sb.delete(sb.length()-1, sb.length());

        return sb.toString();
    }

    /*
     * Return: Hyperperiod of this taskset
     */
    public int hyperperiod() {
        if (nbrTasks() == 0) return 0;

        int[] periods = new int[tasks.size()];
        int i = 0;
        for (Task task : tasks) {
            periods[i] = task.period();
            i++;
        }
        return Util.lcm(periods);
    }

    /*
     * Return: Optional<Task> with priority i or empty optional if it doesn't exist.
     * Throws NoSuchElementException if i is illegal
     */
    public Optional<Task> getTask(int i) {
        if (i == 0) return Optional.of(idle);
        return Optional.ofNullable(tasks.get(i-1));
    }

    /*
     * Return: Priority set
     */
    public Set<Integer> getPriorities() {
        Set<Integer> set = new TreeSet<>();
        for (int i = 1; i <= tasks.size(); i++) set.add(i);
        return set;
    }

    /*
     * Return: number of tasks in taskset
     */
    public int nbrTasks() { return tasks.size(); }

    /*
     * Return: Total taskset utilization()
     */
    public double utilization() {
        double u = 0;
        for (Task task : tasks) u += task.utilization();
        return u;
    }

    /*
     * Return: task with priority i:s utilization
     * or 0 if doesn't exist
     */
    public double utilization(int i) {
        if (i == 0) return idle.utilization();
        return (tasks.size() >= i) ? tasks.get(i-1).utilization() : 0;
    }

    /*
     * Saves the taskset in file
     */
    public void save(String path) throws FileNotFoundException {
        TasksetPrintStream writer = null;
        try {
            writer = new TasksetPrintStream(path);
        } catch (FileNotFoundException e) {
            throw e;
        }

        writer.save(tasks);
        setChanged();
        notifyObservers();
    }

    /*
     * Loads the taskset in file
     */
    public void load(String path) throws FileNotFoundException {
        TasksetBufferedReader reader = null;
        try {
            reader = new TasksetBufferedReader(path);
        } catch (FileNotFoundException e) {
            throw e;
        }

        tasks.clear();
        try {
            reader.load(this);
        } catch (Exception e) {
            return;
        }

        setChanged();
        notifyObservers();
    }
}
